#include "hpp/localstepper/task/feet-area-task.hh"
#include "hpp/localstepper/centipede.hh"
#include "hpp/localstepper/link.hh"
#include "hpp/hik/linear-system.hh"
#include "hpp/gik/robot/standing-robot.hh"
#include "robotDynamics/jrlHumanoidDynamicRobot.h"

ChppLosFeetAreaTask::ChppLosFeetAreaTask ( ChppLosCentipede& inCentipede, ChppLosLink& inLink ) :ChppLosTask ( inCentipede )
{
    attLink = &inLink;

    attConstraint->vector().resize ( 10 );
    attConstraint->matrix().resize ( 10, attCentipede->numberDof() );
    attConstraint->matrix().clear();
    attConstraint->vector().clear();
    attValue.resize ( 10 );

    attStepXi = 0.02;
    attStepDi = 0.05;
    attStepDs = 0.02;

    attThetaXi = 0.08;
    attThetaDi = 0.20;
    attThetaDs = 0.08;
}

void ChppLosFeetAreaTask::setRotationMargins ( double inDi, double inDs, double inXi )
{
    attThetaXi = inXi;
    attThetaDs = inDs;
    attThetaDi = inDi;
}


void ChppLosFeetAreaTask::setTranslationMargins ( double inDi, double inDs, double inXi )
{
    attStepXi = inXi;
    attStepDs = inDs;
    attStepDi = inDi;
}

void ChppLosFeetAreaTask::computeDifferentialSystem()
{
    resize();
    {
        const ChppLosFootprintArea& footprintArea = attLink->leftStepArea();
        const ChppLosJacobianPack& jp = attLink->jacobianPack();
        unsigned int startRank = attLink->rank();

        double coord;

        coord = jp.LinRy;
        if ( coord > footprintArea.maxY - attStepDi )
            attConstraint->vector() ( 0 ) = attStepXi* ( coord - footprintArea.maxY + attStepDs ) / ( attStepDs - attStepDi );
        else
            attConstraint->vector() ( 0 ) = attStepXi;
        for ( unsigned int i=0;i<attLink->numberDof();i++ )
            attConstraint->matrix() ( 0,startRank+i ) = jp.jacLinR ( 1,i );

        coord = jp.LinRx;
        if ( coord > footprintArea.maxX - attStepDi )
            attConstraint->vector() ( 1 ) = attStepXi* ( coord - footprintArea.maxX + attStepDs ) / ( attStepDs - attStepDi );
        else
            attConstraint->vector() ( 1 ) = attStepXi;
        for ( unsigned int i=0;i<attLink->numberDof();i++ )
            attConstraint->matrix() ( 1,startRank+i ) = jp.jacLinR ( 0,i );


        if ( coord < footprintArea.minX + attStepDi )
            attConstraint->vector() ( 2 ) = attStepXi* ( coord - footprintArea.minX - attStepDs ) / ( attStepDi - attStepDs );
        else
            attConstraint->vector() ( 2 ) = attStepXi;
        for ( unsigned int i=0;i<attLink->numberDof();i++ )
            attConstraint->matrix() ( 2,startRank+i ) = -jp.jacLinR ( 0,i );


        coord = jp.LinRth;
        if ( coord > footprintArea.maxTheta - attThetaDi )
            attConstraint->vector() ( 3 ) = attThetaXi* ( coord - footprintArea.maxTheta + attThetaDs ) / ( attThetaDs - attThetaDi );
        else
            attConstraint->vector() ( 3 ) = attThetaXi;
        for ( unsigned int i=0;i<attLink->numberDof();i++ )
            attConstraint->matrix() ( 3,startRank+i ) = jp.jacLYawinR ( i );


        if ( coord < footprintArea.minTheta + attThetaDi )
            attConstraint->vector() ( 4 ) = attThetaXi* ( coord - footprintArea.minTheta - attThetaDs ) / ( attThetaDi - attThetaDs );
        else
            attConstraint->vector() ( 4 ) = attThetaXi;
        for ( unsigned int i=0;i<attLink->numberDof();i++ )
            attConstraint->matrix() ( 4,startRank+i ) = -jp.jacLYawinR ( i );
    }

    {
        const ChppLosFootprintArea& footprintArea = attLink->rightStepArea();
        const ChppLosJacobianPack& jp = attLink->jacobianPack();
        unsigned int startRank = attLink->rank();

        double coord;

        coord = jp.RinLy;
        if ( coord < footprintArea.minY + attStepDi )
            attConstraint->vector() ( 5 ) = attStepXi* ( coord - footprintArea.minY - attStepDs ) / ( attStepDi - attStepDs );
        else
            attConstraint->vector() ( 5 ) = attStepXi;
        for ( unsigned int i=0;i<attLink->numberDof();i++ )
            attConstraint->matrix() ( 5,startRank+i ) = -jp.jacRinL ( 1,i );

        coord = jp.RinLx;
        if ( coord > footprintArea.maxX - attStepDi )
            attConstraint->vector() ( 6 ) = attStepXi* ( coord - footprintArea.maxX + attStepDs ) / ( attStepDs - attStepDi );
        else
            attConstraint->vector() ( 6 ) = attStepXi;
        for ( unsigned int i=0;i<attLink->numberDof();i++ )
            attConstraint->matrix() ( 6,startRank+i ) = jp.jacRinL ( 0,i );


        if ( coord < footprintArea.minX + attStepDi )
            attConstraint->vector() ( 7 ) = attStepXi* ( coord - footprintArea.minX - attStepDs ) / ( attStepDi - attStepDs );
        else
            attConstraint->vector() ( 7 ) = attStepXi;
        for ( unsigned int i=0;i<attLink->numberDof();i++ )
            attConstraint->matrix() ( 7,startRank+i ) = -jp.jacRinL ( 0,i );


        coord = jp.RinLth;
        if ( coord > footprintArea.maxTheta - attThetaDi )
            attConstraint->vector() ( 8 ) = attThetaXi* ( coord - footprintArea.maxTheta + attThetaDs ) / ( attThetaDs - attThetaDi );
        else
            attConstraint->vector() ( 8 ) = attThetaXi;
        for ( unsigned int i=0;i<attLink->numberDof();i++ )
            attConstraint->matrix() ( 8,startRank+i ) = jp.jacRYawinL ( i );


        if ( coord < footprintArea.minTheta + attThetaDi )
            attConstraint->vector() ( 9 ) = attThetaXi* ( coord - footprintArea.minTheta - attThetaDs ) / ( attThetaDi - attThetaDs );
        else
            attConstraint->vector() ( 9 ) = attThetaXi;
        for ( unsigned int i=0;i<attLink->numberDof();i++ )
            attConstraint->matrix() ( 9,startRank+i ) = -jp.jacRYawinL ( i );
    }
}

const vectorN& ChppLosFeetAreaTask::currentValue()
{
    attValue.clear();
    const ChppLosJacobianPack& jp = attLink->jacobianPack();

    {
        const ChppLosFootprintArea& fa = attLink->leftStepArea();

        if ( jp.LinRy > fa.maxY - attStepDs )
            attValue ( 0 ) = fa.maxY - attStepDs - jp.LinRy;
        if ( jp.LinRx > fa.maxX - attStepDs )
            attValue ( 1 ) = fa.maxX - attStepDs - jp.LinRx;
        if ( jp.LinRx < fa.minX + attStepDs )
            attValue ( 2 ) = fa.minX + attStepDs - jp.LinRx;
        if ( jp.LinRth > fa.maxTheta - attThetaDs )
            attValue ( 3 ) = fa.maxTheta - attThetaDs - jp.LinRth;
        if ( jp.LinRth < fa.minTheta + attThetaDs )
            attValue ( 4 ) = fa.minTheta + attThetaDs - jp.LinRth;
    }

    {
        const ChppLosFootprintArea& fa = attLink->rightStepArea();
        if ( jp.RinLy < fa.minY + attStepDs )
            attValue ( 5 ) = fa.minY + attStepDs - jp.RinLy;
        if ( jp.RinLx > fa.maxX - attStepDs )
            attValue ( 6 ) = fa.maxX - attStepDs - jp.RinLx;
        if ( jp.RinLx < fa.minX + attStepDs )
            attValue ( 7 ) = fa.minX + attStepDs - jp.RinLx;
        if ( jp.RinLth > fa.maxTheta - attThetaDs )
            attValue ( 8 ) = fa.maxTheta - attThetaDs - jp.RinLth;
        if ( jp.RinLth < fa.minTheta + attThetaDs )
            attValue ( 9 ) = fa.minTheta + attThetaDs - jp.RinLth;
    }

    return attValue;
}


