#include "boost/numeric/ublas/vector_proxy.hpp"
#include "boost/numeric/ublas/matrix_proxy.hpp"
#include "hpp/localstepper/task/parallel-task.hh"
#include "hpp/localstepper/centipede.hh"
#include "hpp/localstepper/link.hh"
#include "hpp/gik/tools.hh"
#include "hpp/hik/linear-system.hh"
#include "hpp/gik/robot/standing-robot.hh"
#include "robotDynamics/jrlHumanoidDynamicRobot.h"
using namespace boost::numeric::ublas;

#define M3_IJ MAL_S3x3_MATRIX_ACCESS_I_J
#define M4_IJ MAL_S4x4_MATRIX_ACCESS_I_J

void ChppLosParallelTask::equivAsymMat3 ( const vector3d& inVec, matrixNxP& outMat )
{
    outMat ( 0,0 ) = 0;
    outMat ( 0,1 ) = -inVec ( 2 );
    outMat ( 0,2 ) = inVec ( 1 );
    outMat ( 1,0 ) = inVec ( 2 );
    outMat ( 1,1 ) = 0;
    outMat ( 1,2 ) = -inVec ( 0 );
    outMat ( 2,0 ) = -inVec ( 1 );
    outMat ( 2,1 ) = inVec ( 0 );
    outMat ( 2,2 ) = 0;
}

void ChppLosParallelTask::crossProd3 ( const vector3d& v1, const vector3d& v2, vector3d& retV )
{
    retV ( 0 ) = v1 ( 1 ) *v2 ( 2 ) - v1 ( 2 ) *v2 ( 1 );
    retV ( 1 ) = v1 ( 2 ) *v2 ( 0 ) - v1 ( 0 ) *v2 ( 2 );
    retV ( 2 ) = v1 ( 0 ) *v2 ( 1 ) - v1 ( 1 ) *v2 ( 0 );
}

void ChppLosParallelTask::Rodrigues ( const vector3d& inW, double inAngle, matrixNxP& outRot )
{
    double nor = MAL_S3_VECTOR_NORM ( inW );
    identity_matrix<double> id ( 3 );
    if ( nor < 1e-8 )
        outRot = id;
    else
    {
        vector3d wn = inW / nor;
        equivAsymMat3 ( wn,outRot );
        outRot = id + sin ( inAngle ) *outRot + ( 1-cos ( inAngle ) ) * prod ( outRot,outRot );
    }
}

ChppLosParallelTask::ChppLosParallelTask ( ChppLosCentipede& inCentipede, ChppLosRobotLink& inRobotLink, CjrlJoint& inJoint,const vector3d& inLocalVector,const vector3d& inTarget ) :ChppLosTaskOnJoint(inCentipede,inJoint)
{
    attRobotLink = &inRobotLink;
    attLocalVector = inLocalVector;
    attTargetVector = inTarget;

    attConstraint->matrix().resize ( 3,attCentipede->numberDof() );
    attConstraint->matrix().clear();
    attConstraint->vector().resize ( 3 );
    attTempJacobian.resize ( 3, attCentipede->numberDof() );
    attValue.resize ( 1 );
    attMaxVelocity = M_PI/2;
    matOP.resize ( 3,3 );
    matOT.resize ( 3,3 );
    matOTOP.resize ( 3,3 );
    MAL_S3x3_MATRIX_SET_IDENTITY(m3);
}

bool ChppLosParallelTask::resize()
{
    if (ChppLosTask::resize())
    {
    attTempJacobian.resize(3,attCentipede->numberDof());
    return true;
    }
    return false;
}

void ChppLosParallelTask::joint ( CjrlJoint& inJ )
{
    attJoint = &inJ;
}

const vector3d& ChppLosParallelTask::localVector()
{
    return attLocalVector;
}
void ChppLosParallelTask::localVector ( const vector3d& inVec )
{
    attLocalVector = inVec;
    double no = MAL_S3_VECTOR_NORM ( attLocalVector );
    if ( no != 0 )
        attLocalVector = attLocalVector/no;
}

const vector3d& ChppLosParallelTask::targetVector()
{
    return attTargetVector;
}

void ChppLosParallelTask::targetVector ( const vector3d& inVec )
{
    attTargetVector = inVec;
    double no = MAL_S3_VECTOR_NORM ( attTargetVector );
    if ( no != 0 )
        attTargetVector = attTargetVector/no;
}

void ChppLosParallelTask::maxVelocityNorm ( double inMV )
{
    attMaxVelocity = (inMV>0.0)?inMV:0.0;
}

const vectorN& ChppLosParallelTask::currentValue()
{
    for ( unsigned int i=0;i<3;i++ )
        for ( unsigned int j=0;j<3;j++ )
            M3_IJ(m3, i, j ) = M4_IJ(attJoint->currentTransformation(), i, j );
    vector3d locVec = m3*attLocalVector;
    vector3d cross;
    crossProd3 ( attTargetVector,locVec,cross );
    double sv = MAL_S3_VECTOR_NORM ( cross );
    double cv = locVec ( 0 ) *attTargetVector ( 0 ) + locVec ( 1 ) *attTargetVector ( 1 ) + locVec ( 2 ) *attTargetVector ( 2 );
    attValue ( 0 ) = atan2 ( sv,cv );
    return attValue;
}

void ChppLosParallelTask::computeDifferentialSystem()
{
    resize();
    //task vector
    for ( unsigned int i=0;i<3;i++ )
        for ( unsigned int j=0;j<3;j++ )
            M3_IJ(m3, i, j ) = M4_IJ(attJoint->currentTransformation(), i, j );
    vector3d locVec = m3*attLocalVector;
    vector3d cross;
    crossProd3 ( attTargetVector,locVec,cross );
    attConstraint->vector() ( 0 ) = cross ( 0 );
    attConstraint->vector() ( 1 ) = cross ( 1 );
    attConstraint->vector() ( 2 ) = cross ( 2 );
    double normcross = MAL_S3_VECTOR_NORM ( cross );
    double sv = normcross;
    double cv = locVec ( 0 ) *attTargetVector ( 0 ) + locVec ( 1 ) *attTargetVector ( 1 ) + locVec ( 2 ) *attTargetVector ( 2 );
    double angl = atan2 ( sv,cv );
    vector3d worVec = attTargetVector;
    if ( angl > M_PI/2 - 1e-2 )
    {
        matrixNxP rot ( 3,3 );
        matrix3d rot3;
        Rodrigues ( cross, angl/2, rot );
        ChppGikTools::UblastoMatrix3 ( rot, rot3 );
        worVec = rot3 * worVec;
        crossProd3 ( worVec,locVec,cross );
        attConstraint->vector() ( 0 ) = cross ( 0 );
        attConstraint->vector() ( 1 ) = cross ( 1 );
        attConstraint->vector() ( 2 ) = cross ( 2 );
        normcross = MAL_S3_VECTOR_NORM ( cross );
    }

    if ( normcross > attMaxVelocity )
        attConstraint->vector() = attMaxVelocity*attConstraint->vector() /normcross;

    //Body jacobian
    attTempJacobian.clear();
    attRobotLink->standingRobot().robot()->getOrientationJacobian ( *attRobotLink->rootJoint(),*attJoint, attTempJacobian, attRobotLink->rank(), false );
    ChppLosLink* plink = attRobotLink->parent();
    while ( plink!=0 )
    {
        plink->getOrientationJacobian ( attTempJacobian,plink->rank() );
        plink = plink->parent();
    }

    //task jacobian
    equivAsymMat3 ( worVec,matOT );
    equivAsymMat3 ( locVec,matOP );
    noalias ( matOTOP ) = prod ( matOT,matOP );
    noalias ( attConstraint->matrix() ) = prod ( matOTOP,attTempJacobian ) ;
}

